<?php


namespace App\Http\Controllers;


use App\models\Lead;
use App\models\Member;
use App\TraitLibraries\ModelHelper;
use App\User;
use Carbon\Carbon;
use Firebase\JWT\JWT;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Hash;

use Illuminate\Support\Facades\Validator;

class AuthController extends Controller
{
    /**
     * The request instance.
     *
     * @var \Illuminate\Http\Request
     */
    private $request;

    /**
     * Create a new controller instance.
     *
     * @param \Illuminate\Http\Request $request
     * @return void
     */
    public function __construct(Request $request)
    {
        $this->request = $request;
    }


    protected function jwtUsers(User $user)
    {
        $payload = [
            'iss' => "batmobile", // Issuer of the token
            'sub' => $user, // Subject of the token
            'iat' => time(), // Time when JWT was issued.
            'exp' => time() + 6000 * 6000 // Expiration time
        ];
        // As you can see we are passing `JWT_SECRET` as the second parameter that will
        // be used to decode the token in the future.
        return JWT::encode($payload, env('JWT_SECRET'));
    }


    public function getMember(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'email' => 'email',
            'name' => 'required',
            'phone' => 'required|regex:/(03)[0-9]{9}/|min:11|max:11'
        ]);
        try {
            $createNew = $request->get('createme', 0);
            if ($validator->fails()) {
                return $this->sendResult('Invalid member details', config('codes.REQUEST_VALIDATAION_ERROR'), $validator->errors());
            }
            // Find the user by email
            $user = Member::where('phone', $this->request->input('phone'))->first();

            if ($createNew && !$user) {
                $user = new Member(
                    [
                        'email' => $request->get('email'),
                        'name' => $request->get('name'),
                        'phone' => $request->get('phone'),
                        'reference' => 'Fitness'
                    ]
                );
                $user->save();
            } elseif (!$user) {
                throw new \Exception("No Member found against {$request->get('email')}");
            }
            $member = [
                'member_id' => $user->id,
                'name' => $user->name,
                'phone' => $user->phone,
                'email' => $user->email,
            ];

            return $this->sendResult("Welcome {$user['name']}", config('codes.SUCCESS'), compact('member'));

        } catch (\Exception $exception) {
            return $this->sendResult($exception->getMessage(), config('codes.RESOURCE_NOT_FOUND'), []);
        }

    }

    public function authenticate(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'email' => 'required|email',
            'password' => 'required'
        ]);
        try {
            if ($validator->fails()) {
                return $this->sendResult('Invalid email and password', config('codes.REQUEST_VALIDATAION_ERROR'), $validator->errors());
            }
            // Find the user by email
            $user = User::where('email', $this->request->input('email'))->first();

            if (!$user) {
                throw new \Exception("No user found against {$request->get('email')}");
            }

            if (!Hash::check($this->request->input('password'), $user->password)) {
                throw new \Exception("Invalid password");
            }

            $token = $this->payLoad($user);
            $user = $user->toArray();

            return $this->sendResult("Welcome {$user['name']}", config('codes.SUCCESS'), compact('user', 'token'));

        } catch (\Exception $exception) {
            return $this->sendResult($exception->getMessage(), config('codes.RESOURCE_NOT_FOUND'), []);
        }

    }

    public function handleLogin(Request $request)
    {
        $validator = Validator::make($request->all(), [
            "uid" => "required",
        ]);
        try {
            if ($validator->fails()) {
                return $this->sendResult('Error In Registering member', config('codes.REQUEST_VALIDATAION_ERROR'), $validator->errors());
            }

            $auth = app('firebase.auth');
            $apiUser = $auth->getUser($request->get('uid'));

            if (!empty($apiUser->phoneNumber)) {
                $mytime = Carbon::now();
                $phoneNumber = ModelHelper::phoneLocalFormat($apiUser->phoneNumber);

                $member = Member::where("phone", $phoneNumber)->first();

                if (empty($member)) {
                    $member = new Member([
                        "phone" => $phoneNumber,
                        "uid" => $apiUser->uid,
                        "mobile_verified_at" => $mytime,
                        "created_at" => $mytime,
                        "updated_at" => $mytime,
                        'password' => Hash::make($apiUser->phoneNumber)
                    ]);

                    if (!$member->save())
                        return $this->sendResult(config('codes.DATA_NOT_SAVE_MSG'), config('codes.DATA_NOT_SAVE'), [], $request);

                    $member = Member::findOrFail($member->id);

                    if ($request->get("fcm_token"))
                        $member->updateFcmToken($request->get("fcm_token"));
                    Lead::create([
                        "phone" => $phoneNumber,
                        "type" => "MOBILE"
                    ]);

                }
                $token = $this->payLoad($member);
                $member = $member->toArray();

                return $this->sendResult("Welcome {$member['name']}", config('codes.SUCCESS'), compact('member', 'token'), $request);
            }
            return $this->sendResult("Member not found", config('codes.DATA_NOT_FOUND'), [], $request);

        } catch (\Exception $exception) {
            return $this->sendResult($exception->getMessage(), config('codes.RESOURCE_NOT_FOUND'), [], $request);
        }

    }

    public function handleLoginOpen(Request $request)
    {
        $validator = Validator::make($request->all(), [
            "phone" => "required",
        ]);

        try {

            if ($validator->fails()) {
                return $this->sendResult('Error In Registering member', config('codes.REQUEST_VALIDATAION_ERROR'), $validator->errors());
            }

            $phoneNumber = $request->get("phone");
            $countryCode = '0'; // Replace with known country code of user.
            $internationalNumber = preg_replace('/^#00/', $countryCode, $phoneNumber);
            $member = Member::where("phone", $internationalNumber)->first();

            if (empty($member)) {
                return $this->sendResult("Member not found", config('codes.DATA_NOT_FOUND'), [], $request);
            }

            $token = $this->payLoad($member);
            if ($request->get("fcm_token"))
                $member->updateFcmToken();
            $member = $member->toArray();

            return $this->sendResult("Welcome {$member['name']}", config('codes.SUCCESS'), compact('member', 'token'), $request);

        } catch (\Exception $exception) {
            return $this->sendResult($exception->getMessage(), config('codes.RESOURCE_NOT_FOUND'), [], $request);
        }

    }


}
